#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <limits.h>
#include <setjmp.h>
#include <locale.h>

#include "config.h"
#include "imapfilter.h"
#include "version.h"
#include "account.h"
#include "filter.h"
#include "buffer.h"

#if defined SSL_TLS || defined ENCRYPTED_PASSWORDS || defined CRAM_MD5
#include <openssl/opensslv.h>
#include <openssl/crypto.h>
#endif


extern account_t *accounts;
extern filter_t *filters;
extern buffer_t ibuf, obuf;

unsigned int options;		/* Program options. */
unsigned int flags = 0;		/* Program flags. */
unsigned int interval = 0;	/* Poll at the specified interval. */
conn_t connpri, connaux;	/* Primary and auxiliary IMAP connection. */

char logfile[PATH_MAX];		/* Log file. */
char *home = NULL;		/* User's home directory. */
char charset[CHARSET_LEN];	/* Charset for IMAP SEARCH requests. */

jmp_buf acctloop;		/* Non-local exit in case of network error. */


void usage(void);
void version(void);


/*
 * IMAPFilter: an IMAP mail filtering utility.
 */
int
main(int argc, char *argv[])
{
	int c, r, f;
	pid_t pid;
	char *conffile;		/* Configuration file. */
	account_t *ca;		/* Current account. */
	mbox_t *cm;		/* Current mailbox. */

	setlocale(LC_ALL, "");

	f = 0;
	home = getenv("HOME");
	options = (OPTION_DETAILS_NORMAL | OPTION_NAMESPACE | OPTION_PEEK);
	*charset = 0;
	*logfile = 0;
	conffile = NULL;
	connpri.sock = connaux.sock = -1;

	while ((c = getopt(argc, argv, "c:d:hkl:"
#ifdef ENCRYPTED_PASSWORDS
		    "p"
#endif
		    "qvV")) != -1) {
		switch (c) {
		case 'c':
			conffile = optarg;
			break;
		case 'd':
			options |= OPTION_DAEMON_MODE;
			errno = 0;
			interval = strtoul(optarg, NULL, 10);
			if (errno)
				interval = 0;
			break;
		case 'h':
			usage();
			exit(ERROR_UNDEFINED);
			break;
		case 'k':
			kill_imapfilter();
			break;
		case 'l':
			strncat(logfile, optarg, PATH_MAX - 1);
			break;
#ifdef ENCRYPTED_PASSWORDS
		case 'p':
			options |= OPTION_PASSWORD_EDITOR;
			break;
#endif
		case 'q':
			options &= OPTION_DETAILS_CLEAR;
			options |= OPTION_DETAILS_QUIET;
			break;
		case 'v':
			options &= OPTION_DETAILS_CLEAR;
			options |= OPTION_DETAILS_VERBOSE;
			break;
		case 'V':
			version();
			exit(ERROR_UNDEFINED);
			break;
		default:
			usage();
			exit(ERROR_UNDEFINED);
			break;
		}
	}

	create_homedir();

	lockfile_check();
	lockfile_create();

#ifndef DEBUG
	corefile_disable();
#endif

	tty_store();
	catch_signals();

	read_config(conffile);

#ifdef ENCRYPTED_PASSWORDS
	read_passwords();

	if ((options & OPTION_PASSWORD_EDITOR)) {
		password_editor();

		secmem_clear();
		lockfile_remove();

		exit(0);
	}
#endif

	open_logfile();

	init_buffer(&ibuf);
	init_buffer(&obuf);

	if (options & OPTION_DAEMON_MODE) {
		f = 1;
		options &= OPTION_DETAILS_CLEAR;
		options |= OPTION_DETAILS_QUIET;
	}
	do {
		for (ca = accounts; ca != NULL; ca = ca->next) {

			if (setjmp(acctloop))
				continue;

			if (init_connection(&connpri, ca->server, ca->port,
				ca->ssl))
				continue;

			r = greeting_response(&connpri);

#ifdef DEBUG
			test(&connpri);
#endif

			if (check_capabilities(&connpri))
				continue;

#ifdef SSL_TLS
			if (ca->ssl == SSL_DISABLED &&
			    connpri.caps & CAPABILITY_STARTTLS)
				if (negotiate_tls(&connpri) == RESPONSE_OK)
					check_capabilities(&connpri);
#endif

			log_info(LOG_ACCOUNT, ca->key);

			if (r != RESPONSE_PREAUTH) {
				if (ca->passwdattr == PASSWORD_NONE) {
					printf("Enter password for %s@%s: ",
					    ca->username, ca->server);
					get_password(ca->password, PASSWORD_LEN);
					ca->passwdattr = PASSWORD_PLAIN;
				}
#ifdef CRAM_MD5
				if (connpri.caps & CAPABILITY_AUTH_CRAM_MD5)
					r = auth_cram_md5(&connpri,
					    ca->username, ca->password);
				else
#endif
					r = login(&connpri, ca->username,
					    ca->password);

				if (r == RESPONSE_NO) {
					error("username %s or password rejected "
					    "at %s\n", ca->username, ca->server);
					continue;
				}
			}
			check_namespace(&connpri);

			for (cm = ca->mboxes; cm != NULL; cm = cm->next)
				if (*cm->filters == NULL)
					mailbox_status(&connpri, cm->name);
				else if (!select_mailbox(&connpri, cm->name)) {
					apply_filters(cm->name, cm->filters);
					close_mailbox(&connpri);
				}
			logout(&connpri);

			close_connection(&connpri);
		}

		/* Fork if in daemon mode. */
		if (f) {
			f = 0;
			pid = fork();
			switch (pid) {
			case -1:
				fatal(ERROR_FORK, "forking; %s\n",
				    strerror(errno));
				break;
			case 0:
				lockfile_create();
				corefile_disable();
				flags |= FLAG_DAEMON_MODE;
				break;
			default:
				secmem_clear();
				close_logfile();
				exit(0);
				break;
			}
		}
		if (options & OPTION_DAEMON_MODE &&
		    flags & FLAG_SIGUSR1_RECEIVED) {
			reread_config(conffile);
			continue;
		}
		if (interval)
			sleep(interval);
	} while (options & OPTION_DAEMON_MODE && interval);

	secmem_clear();
	close_logfile();

	lockfile_remove();

	exit(0);
}


/*
 * Print a very brief usage message.
 */
void
usage(void)
{
	fprintf(stderr,
	    "usage: imapfilter [-hk"
#ifdef ENCRYPTED_PASSWORDS
	    "p"
#endif
	    "qvV] [-c configfile] [-d interval] [-l logfile]\n");
}


/*
 * Print program's version, and if it is built in, OpenSSL's version number.
 */
void
version(void)
{
	fprintf(stderr, "IMAPFilter %s"
#if defined SSL_TLS || defined ENCRYPTED_PASSWORDS || defined CRAM_MD5
	    ", OpenSSL 0x%8.8lx"
#endif
	    "\n", IMAPFILTER_VERSION
#if defined SSL_TLS || defined ENCRYPTED_PASSWORDS || defined CRAM_MD5
	    ,SSLeay()
#endif
	);
}
