#include <stdio.h>
#include <string.h>

#include "config.h"
#include "imapfilter.h"
#include "account.h"
#include "filter.h"


extern conn_t connpri, connaux;
extern char charset[CHARSET_LEN];


int match_filter(filter_t * filter, char **mesgs);

char *generate_filter_and(mask_t * mask, unsigned int masknum, unsigned int masklen);
char *generate_filter_or(mask_t * mask, unsigned int masknum, unsigned int masklen);
char *normalize_mask(char *body);
void convert_date(char *body);

void empty_fifo(mask_t ** mfifo);
void queue_fifo(mask_t ** mfifo, mask_t * mask);
mask_t *dequeue_fifo(mask_t ** mfifo);


/*
 * Match and apply filters assigned to a mailbox.
 */
int
apply_filters(char *mbox, filter_t ** filters)
{
	int i;
	char *mesgs;

	for (i = 0; filters[i] != NULL; i++) {
		mesgs = NULL;

		if (match_filter(filters[i], &mesgs))
			continue;

		log_info(LOG_FILTER, filters[i]->key);

		apply_action(mbox, mesgs, &(filters[i]->action.type),
		    filters[i]->action.raccount, filters[i]->action.destmbox,
		    &filters[i]->action.msgflags, filters[i]->action.args);

		xfree(mesgs);
	}

	return 0;
}


/*
 * Generate the search request by the masks of the filter and try to
 * match the generated filter.
 */
int
match_filter(filter_t * filter, char **mesgs)
{
	char *search;

	if (filter->mode == FILTER_MODE_OR)
		search = generate_filter_or(filter->masks, filter->masknum,
		    filter->masklen);
	else
		search = generate_filter_and(filter->masks, filter->masknum,
		    filter->masklen);

	search_response(&connpri, imap_search(&connpri, charset, search),
	    mesgs);

	xfree(search);

	if (*mesgs == '\0')
		return 1;

	return 0;
}


/*
 * Generate the filter search command from the masks, assuming that
 * masks are AND-ed.
 */
char *
generate_filter_and(mask_t * mask, unsigned int masknum,
    unsigned int masklen)
{
	const unsigned int searchbuf = masklen + masknum * 6 + 8;
	unsigned int len;
	char *search;
	mask_t *tmp;

	len = 0;

	search = (char *)xmalloc(sizeof(char) * searchbuf);
	search[0] = '\0';

	tmp = mask;
	if (tmp == NULL) {
		strncat(search, "ALL ", searchbuf - len - 1);
		len += 4;
	} else
		while ((tmp = tmp->next) != NULL) {
			if (tmp->type != MASK_TYPE_OR) {
				strncat(search, "ALL ", searchbuf - len - 1);
				len += 4;
				break;
			}
		}

	tmp = NULL;
	while (mask != NULL) {
		tmp = mask;
		mask = mask->next;

		if (mask != NULL && mask->type == MASK_TYPE_OR) {
			strncat(search, "OR (", searchbuf - len - 1);
			len += 4;

			strncat(search, normalize_mask(tmp->body),
			    searchbuf - len - 1);
			len = strlen(search);
			search[len] = ' ';
			search[++len] = '\0';

			search[len - 1] = ')';
			search[len] = ' ';
			search[++len] = '\0';

			if (mask->next == NULL ||
			    mask->next->type != MASK_TYPE_OR) {
				search[len] = '(';
				search[++len] = '\0';
				strncat(search, normalize_mask(mask->body),
				    searchbuf - len - 1);
				len = strlen(search);
				search[len] = ')';
				search[++len] = ' ';
				search[++len] = '\0';
				mask = mask->next;
			}
		} else {
			strncat(search, normalize_mask(tmp->body),
			    searchbuf - len - 1);
			len = strlen(search);
			search[len] = ' ';
			search[++len] = '\0';
		}
	}

	search[len - 1] = '\0';

	return search;
}


/*
 * Generate the filter search command from the masks, assuming that
 * masks are OR-ed.
 */
char *
generate_filter_or(mask_t * mask, unsigned int masknum,
    unsigned int masklen)
{
	const unsigned int searchbuf = masklen + masknum * 6 + 8;
	unsigned int len;
	char *search;
	mask_t **mfifo;		/* Mailbox FIFO queue. */
	mask_t *mf;		/* Mask returned from FIFO. */

	len = 0;

	search = (char *)xmalloc(sizeof(char) * searchbuf);
	mfifo = (mask_t **) xmalloc(sizeof(mask_t *) * (masknum + 1));

	search[0] = '\0';
	empty_fifo(mfifo);

	strncat(search, "ALL ", searchbuf - len - 1);
	len += 4;

	while (mask != NULL) {
		queue_fifo(mfifo, mask);
		mask = mask->next;

		while (mask != NULL && mask->type == MASK_TYPE_AND) {
			queue_fifo(mfifo, mask);
			mask = mask->next;
		}

		if (mask != NULL) {
			if (len == 4 && search[0] == 'A') {
				search[0] = '\0';
				len = 0;
			}
			strncat(search, "OR ", searchbuf - len - 1);
			len += 3;
		}
		if (search[0] != 'A') {
			search[len] = '(';
			search[++len] = '\0';
		}
		while ((mf = dequeue_fifo(mfifo)) != NULL) {
			strncat(search, normalize_mask(mf->body), searchbuf - len - 1);
			len = strlen(search);
			search[len] = ' ';
			search[++len] = '\0';
		}

		if (strchr(search, '(')) {
			search[len - 1] = ')';
			search[len] = ' ';
			search[++len] = '\0';
		}
		empty_fifo(mfifo);
	}

	search[len - 1] = '\0';

	xfree(mfifo);

	return search;
}


/*
 * Make any necessary conversions to the mask body before sending it to server.
 */
char *
normalize_mask(char *body)
{
	static char nb[MASK_BODY_LEN];


	if (strstr(body, "OLDER") ||
	    strstr(body, "NEWER")) {
		xstrncpy(nb, body, MASK_BODY_LEN - 1);
		convert_date(nb);

		return nb;
	}
	return body;
}


/*
 * Converts masks related to date filtering, because IMAP servers do not
 * understand for example "OLDER 3", but "BEFORE 18-Oct-2001" (if
 * hypothetically current date was 21-Oct-2001).
 */
void
convert_date(char *body)
{
	char *cp, *c;
	char s[16];
	time_t te;
	struct tm *tl;

	cp = xstrdup(body);
	body[0] = '\0';

	if (strstr(cp, "NOT"))
		strncat(body, "NOT ", 4);

	if ((c = strstr(cp, "OLDER")))
		strncat(body, "BEFORE ", 7);
	else if ((c = strstr(cp, "NEWER")))
		strncat(body, "SINCE ", 6);

	c += 6;

	te = time(NULL) - (time_t) (strtoul(c, NULL, 10) * 24 * 60 * 60);
	tl = localtime(&te);

	if (strftime(s, 15, "%d-%b-%Y", tl))
		strncat(body, s, 15);

	xfree(cp);
}


/*
 * Empty the FIFO inventory.
 */
void
empty_fifo(mask_t ** mfifo)
{
	mfifo[0] = NULL;

	queue_fifo(NULL, NULL);
	dequeue_fifo(NULL);
}


/*
 * Add item to FIFO inventory.
 */
void
queue_fifo(mask_t ** mfifo, mask_t * mask)
{
	static unsigned int i;

	if (mfifo == NULL) {
		i = 0;
		return;
	}
	mfifo[i++] = mask;
	mfifo[i] = NULL;
}


/*
 * Get next item from FIFO inventory.
 */
mask_t *
dequeue_fifo(mask_t ** mfifo)
{
	static unsigned int j;

	if (mfifo == NULL) {
		j = 0;
		return NULL;
	}
	return mfifo[j++];
}
